//*****************************************************************************
//              \ /
// No Copyright (c) 2003
//              / \
// No rights reserved. This material contains published,
// non-copyrighted work, which includes non-confidential and
// non-proprietary information of Kurt Santel.
//
// Source: CodeWar.h
//
// Description: Server-side component of CodeWar
//
//*****************************************************************************

#pragma once

class Arena;
#include "Robot.h"
#include "../CodeWarCommon/Thread.hpp"
using namespace Thread;
#include <vector>

class CodeWar
{
public:
    CodeWar(void);
    ~CodeWar(void);

    bool initialize(XType width, YType height, unsigned short port);

    Arena * getArena(void) const { return( m_Arena ); }

protected: // "callback" function

    void acceptConnection(SOCKET s);            // Whenever the ListeningThread
                                                //   accepts a new connection, this
                                                //   function is called.

private: // classes

    ////////////////////////////////////////////////////////////////////
    // An instance of this class will be listening for incoming
    // socket connections.  Each client connection will be handled
    // by a separate SlaveThread (see below).
    class ListeningThread : public CThread
    {
    public:
        ListeningThread(CodeWar * cw, unsigned short port);

        void OnInitInstance();                  // Called before OnRunning()
        void OnRunning(void);                   // Start accepting client connections
        void Stop(void);

    protected:
        bool m_bStop;                           // OnRunning() will continue
                                                //   until m_bStop == true
    private:
        CodeWar * m_CodeWar;                    // Needed when a new client tries
                                                //   to connect
        unsigned short m_Port;
        SOCKET m_ListeningSocket;
    };

    ////////////////////////////////////////////////////////////////////
    // Each SlaveThread will handle the communication with one client.
    // When the client disconnects, the thread will end.
    class SlaveThread : public CThread
    {
    public:
        SlaveThread(CodeWar * cw, SOCKET client);
        void OnRunning(void);                   // Communicate with the client
        void Stop(void);
        bool Stopped(void);                     // If m_bStop == true, WaitUntilTerminate()

    protected:
        bool m_bStop;                           // OnRunning() will continue
                                                //   until m_bStop == true
    public: // Made public so that the ControlThread can access them
        SOCKET m_ClientSocket;
        Robot m_Robot;

    private:
        void registerClient();
        void doTurn(CommandType & cmdType);

    private:
        CodeWar * m_CodeWar;
    };

    ////////////////////////////////////////////////////////////////////
    // This one will clean up terminated SlaveThreads.
    class CleanupThread : public CThread
    {
    public:
        CleanupThread(CodeWar * cw);            // Initialize m_bStop to false
        void OnRunning(void);                   // Cleanup terminated threads
        void Stop(void);

    protected:
        bool m_bStop;                           // OnRunning() will continue
                                                //   until m_bStop == true
    private:
        CodeWar * m_CodeWar;                    // Needed when a new client tries
                                                //   to connect
    };

    ////////////////////////////////////////////////////////////////////
    // This class takes care of giving each client a turn.
    // When a client gets a turn, the server will send all the events
    // and commands to the client, so that the client can react properly.
    class ControlThread : public CThread
    {
    public:
        ControlThread(CodeWar * cw);            // Initialize m_bStop to false
        void OnRunning(void);                   // Give each client a turn to play
        void Stop(void);

    protected:
        bool m_bStop;                           // OnRunning() will continue
                                                //   until m_bStop == true
    private:
        void sendGeneralInfo(SlaveThread * slave, NrType nr);
        void sendCommands(SlaveThread * slave); // Send all the events and commands to the
                                                // client when the client has a turn

    private:
        CodeWar * m_CodeWar;                    // Needed to access list with SlaveThreads
    };

private: // data members

    Arena * m_Arena;                            // Where all the fun takes place

    ListeningThread * m_ListeningThread;        // Listen for incoming connections
    CleanupThread * m_CleanupThread;            // Cleanup terminated threads
    ControlThread * m_ControlThread;            // Give each client a turn

    HANDLE m_SlaveThreadMutex;                  // Mutex to protect m_SlaveThreads

    typedef std::vector<SlaveThread *> SlaveThreadType;
    SlaveThreadType m_SlaveThreads;             // SlaveThreads will be created on the fly,
                                                // i.e. when a client makes a connection.

    bool m_UpdateClients;                       // Every time a client connects, the other
                                                // clients get an update.
};
